"""Module with the definition of the thermo-physical property objects
for *bismuth*."""
from typing import List
from typing import Union
import numpy as np
from scipy.constants import atm
from .tph_common_interface import SaturationVapourPressureInterface
from .tph_common_interface import SurfaceTensionInterface
from .tph_common_interface import DensityInterface
from .tph_common_interface import ThermalExpansionInterface
from .tph_common_interface import SpeedOfSoundInterface
from .tph_common_interface import IsentropicCompressibilityInterface
from .tph_common_interface import SpecificHeatInterface
from .tph_common_interface import SpecificEnthalpyInterface
from .tph_common_interface import DynamicViscosityInterface
from .tph_common_interface import ElectricalResistivityInterface
from .tph_common_interface import ThermalConductivityInterface
from .._decorators import range_warning
from .._commons import BISMUTH_MELTING_TEMPERATURE as T_m0
from .._commons import BISMUTH_BOILING_TEMPERATURE as T_b0


class p_s(SaturationVapourPressureInterface):
    """
    Liquid bismuth *saturation vapour pressure* property class.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *saturation vapour pressure* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            saturation vapour pressure in :math:`[Pa]`
        """
        return 2.67e10 * np.exp(-22858/T)

    def initialization_helper(self,
                              property_value: float) -> Union[None, float]:
        """
        Returns the temperature guess value according to the value
        of the saturation vapour pressure passed as argument.
        It is used by the root finder algorithm.

        Parameters
        ----------
        property_value : float
            saturation vapour pressure in :math:`[Pa]`

        Returns
        -------
        float
            Temperature guess value in :math:`[K]`
        """
        if property_value < 1e-2:
            return 800
        if 1e-2 <= property_value < 1e2:
            return 1200
        return 2000

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return 'sobolev2011'

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the saturation vapour
        pressure correlation function
        """
        return [T_m0, T_b0]

    @property
    def units(self) -> str:
        """
        str : Saturation vapour pressure unit
        """
        return "[Pa]"

    @property
    def long_name(self) -> str:
        """
        str : Saturation vapour pressure long name
        """
        return "saturation vapour pressure"

    @property
    def description(self) -> str:
        """
        str : Saturation vapour pressure description
        """
        return f"Liquid bismuth {self.long_name}"


class sigma(SurfaceTensionInterface):
    """
    Liquid bismuth *surface tension* property class.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *surface tension* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            surface tension in :math:`[N/m]`
        """
        return (420.8 - 0.081*T)*1e-3

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return 'sobolev2011'

    @property
    def description(self) -> str:
        """
        str : Surface tension description
        """
        return f"Liquid bismuth {self.long_name}"


class rho(DensityInterface):
    """
    Liquid bismuth *density* property class.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *density* by applying the property
        correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            density in :math:`[kg/m^3]`
        """
        rho_0 = 10725 - 1.22 * T
        u_s_val = u_s().correlation(T, p)
        alpha_val = alpha().correlation(T, p)
        return rho_0 +\
            (1 / u_s_val / u_s_val
             + T * alpha_val * alpha_val / cp().correlation(T, p)) * (p - atm)

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "imbeni1998"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the density
        correlation function
        """
        return [T_m0, T_b0]

    @property
    def description(self) -> str:
        """
        str : Density description
        """
        return f"Liquid bismuth {self.long_name}"


class alpha(ThermalExpansionInterface):
    """
    Liquid bismuth *thermal expansion coefficient* property class.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *thermal expansion coefficient* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            thermal expansion coefficient in :math:`[1/K]`
        """
        return 1/(8791 - T)

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the thermal expansion
        coefficient correlation function
        """
        return [T_m0, T_b0]

    @property
    def description(self) -> str:
        """
        str : Thermal expansion coefficient description
        """
        return f"Liquid bismuth {self.long_name}"


class u_s(SpeedOfSoundInterface):
    """
    Liquid bismuth *sound velocity* property class.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *sound velocity* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            sound velocity in :math:`[m/s]`
        """
        return 1616 + T * (0.187 - 2.2e-4 * T)

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return 'sobolev2011'

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the sound velocity
        correlation function
        """
        return [T_m0, 1800.0]

    @property
    def description(self) -> str:
        """
        str : Sound velocity description
        """
        return "Sound velocity in liquid bismuth"


class beta_s(IsentropicCompressibilityInterface):
    """
    Liquid bismuth *isentropic compressibility* property class.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *isentropic compressibility* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            isentropic compressibility in :math:`[1/Pa]`
        """
        u_s_val = u_s().correlation(T, p)
        return 1 / (rho().correlation(T, p) * u_s_val * u_s_val)

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the isentropic
        compressibility correlation function
        """
        return [T_m0, 1800.0]

    @property
    def description(self) -> str:
        """
        str : Isentropic compressibility description
        """
        return f"Liquid bismuth {self.long_name}"


class cp(SpecificHeatInterface):
    """
    Liquid bismuth *specific heat capacity* property class.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *specific heat capacity* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            specific heat capacity in :math:`[J/(kg \\cdot K)]`
        """
        return 118.2 + 5.934e-3*T + 7.183e6/T/T

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "imbeni1998"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the specific heat
        capacity correlation function
        """
        return [T_m0, T_b0]

    @property
    def description(self) -> str:
        """
        str : Specific heat capacity description
        """
        return f"Liquid bismuth {self.long_name}"


class h(SpecificEnthalpyInterface):
    """
    Liquid bismuth *specific enthalpy* property class.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *specific enthalpy* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            specific enthalpy in :math:`[J/kg]`
        """
        return T * (118.2 + 2.967e-3 * T) - T_m0 * (118.2 + 2.967e-3 * T_m0)\
            - 7.183e6 * (1 / T - 1 / T_m0)

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return 'sobolev2011'

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the specific enthalpy
        correlation function
        """
        return [T_m0, T_b0]

    @property
    def description(self) -> str:
        """
        str : Specific enthalpy description
        """
        return (f"Liquid bismuth {self.long_name} "
                "(as difference with respect to "
                "the melting point enthalpy)")


class mu(DynamicViscosityInterface):
    """
    Liquid bismuth *dynamic viscosity* property class.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *dynamic viscosity* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            dynamic viscosity in :math:`[Pa \\cdot s]`
        """
        return 4.456e-4*np.exp(780/T)

    @property
    def correlation_name(self) -> str:
        """
        str : name of the correlation
        """
        return "lucas1984b"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the dynamic viscosity
        correlation function
        """
        return [T_m0, 1300.0]

    @property
    def description(self) -> str:
        """
        str : Dynamic viscosity description
        """
        return f"Liquid bismuth {self.long_name}"


class r(ElectricalResistivityInterface):
    """
    Liquid bismuth *electrical resistivity* property class.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *electrical resistivity* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            electrical resistivity in :math:`[Ohm \\cdot m]`
        """
        return (98.96 + 0.0554*T)*1e-8

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the electrical
        resistivity correlation function
        """
        return [545.0, 1423.0]

    @property
    def description(self) -> str:
        """
        str : Electrical resistivity description
        """
        return f"Liquid bismuth {self.long_name}"


class k(ThermalConductivityInterface):
    """
    Liquid bismuth *thermal conductivity* property class.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *thermal conductivity* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            thermal conductivity in :math:`[W/(m \\cdot K)]`
        """
        return 7.34 + 9.5e-3*T

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "touloukian1970b"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the thermal
        conductivity correlation function
        """
        return [T_m0, 1000.0]

    @property
    def description(self) -> str:
        """
        str : Thermal conductivity description
        """
        return f"Liquid bismuth {self.long_name}"



# This test is used to check correct implementation of some lead properties
# comparing the Lead class output with data.dat
import unittest
from scipy.constants import convert_temperature
import numpy
import sys
import os
sys.path.insert(0, os.path.abspath('..'))
from lbh15 import Lead

# Specify the correlation to use for cp
Lead.set_correlation_to_use('cp', 'gurvich1991')

properties_table = numpy.loadtxt("data.dat", comments='#',
                                 skiprows=2, dtype=float)
celsius_array = properties_table[:, 0]
rho_array = properties_table[:, 1]
k_array = properties_table[:, 2]
cp_array = properties_table[:, 3]
mu_array = properties_table[:, 4]
tol = 4


class LeadComparisonTester(unittest.TestCase):

    def test_vs_data(self):
        for i in range(len(celsius_array)):
            liquid_lead = Lead(T=convert_temperature(celsius_array[i],
                                                     'C', 'K'))
            self.assertAlmostEqual(liquid_lead.rho, rho_array[i], tol)
            self.assertAlmostEqual(liquid_lead.k, k_array[i], tol)
            self.assertAlmostEqual(liquid_lead.cp, cp_array[i], tol)
            self.assertAlmostEqual(liquid_lead.mu, mu_array[i], tol)


if __name__ == "__main__":
    unittest.main()



"""Module with the definition of the *diffusivity*
property objects for *lead*."""
from typing import List
import numpy as np
from scipy.constants import atm
from scipy.constants import R
from ..interface import PropertyInterface
from ..tch_common_interface import OxygenDiffusivityInterface
from ..tch_common_interface import IronDiffusivityInterface
from ..._decorators import range_warning


class LeadOxygenDiffusivityInterface(OxygenDiffusivityInterface):
    """
    Liquid lead *Oxygen diffusivity* property abstract class.
    """
    @property
    def description(self) -> str:
        """
        str : Oxygen diffusivity description
        """
        return f"{self.long_name} in liquid lead"


class OxygenDiffusivityArcella1968(LeadOxygenDiffusivityInterface):
    """
    Liquid lead *Oxygen diffusivity* property class
    implementing the correlation by *arcella1968*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen diffusivity* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            diffusivity in :math:`[m^2 / s]`
        """
        return np.exp(-14979 / R / T) * 6.32e-9

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "arcella1968"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen diffusivity
        correlation function
        """
        return [973, 1173]


class OxygenDiffusivityHomna1971(LeadOxygenDiffusivityInterface):
    """
    Liquid lead *Oxygen diffusivity* property class
    implementing the correlation by *homna1971*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen diffusivity* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            diffusivity in :math:`[m^2 / s]`
        """
        return np.exp(-20083 / R / T) * 9.65e-9

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "homna1971"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen diffusivity
        correlation function
        """
        return [1073, 1373]


class OxygenDiffusivitySwzarc1972(LeadOxygenDiffusivityInterface):
    """
    Liquid lead *Oxygen diffusivity* property class
    implementing the correlation by *swzarc1972*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen diffusivity* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            diffusivity in :math:`[m^2 / s]`
        """
        return np.exp(-25942 / R / T) * 1.44e-7

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "swzarc1972"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen diffusivity
        correlation function
        """
        return [1013, 1353]


class OxygenDiffusivityOtsuka1975(LeadOxygenDiffusivityInterface):
    """
    Liquid lead *Oxygen diffusivity* property class
    implementing the correlation by *otsuka1975*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen diffusivity* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            diffusivity in :math:`[m^2 / s]`
        """
        return np.exp(-19497 / R / T) * 1.48e-7

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "otsuka1975"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen diffusivity
        correlation function
        """
        return [1173, 1373]


class OxygenDiffusivityCharle1976(LeadOxygenDiffusivityInterface):
    """
    Liquid lead *Oxygen diffusivity* property class
    implementing the correlation by *charle1976*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen diffusivity* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            diffusivity in :math:`[m^2 / s]`
        """
        return np.exp(-20927 / R / T) * 1.90e-7

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "charle1976"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen diffusivity
        correlation function
        """
        return [1173, 1373]


class OxygenDiffusivityGromov1996(LeadOxygenDiffusivityInterface):
    """
    Liquid lead *Oxygen diffusivity* property class
    implementing the correlation by *gromov1996*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen diffusivity* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            diffusivity in :math:`[m^2 / s]`
        """
        return np.exp(-16158 / R / T) * 6.6e-9

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "gromov1996"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen diffusivity
        correlation function
        """
        return [673, 1273]


class OxygenDiffusivityGanesan2006b(LeadOxygenDiffusivityInterface):
    """
    Liquid lead *Oxygen diffusivity* property class
    implementing the correlation by *ganesan2006b*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen diffusivity* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            diffusivity in :math:`[m^2 / s]`
        """
        return np.exp(-45587 / R / T) * 2.79e-7

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "ganesan2006b"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen diffusivity
        correlation function
        """
        return [823, 1053]


class IronDiffusivity(IronDiffusivityInterface):
    """
    Liquid lead *Iron diffusivity* property class.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Iron diffusivity* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            diffusivity in :math:`[m^2 / s]`
        """
        return np.power(10, - 2.31 - 2295 / T) * 1.0e-4

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Iron diffusivity
        correlation function
        """
        return [973.0, 1273.0]

    @property
    def description(self) -> str:
        """
        str : Iron diffusivity description
        """
        return f"{self.long_name} in liquid lead"


class CobaltDiffusivity(PropertyInterface):
    """
    Liquid lead *Cobalt diffusivity* property class.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Cobalt diffusivity* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            diffusivity in :math:`[m^2 / s]`
        """
        return np.exp(-22154 / R / T) * 4.6e-8

    @property
    def name(self) -> str:
        """
        str : Name of the property
        """
        return "co_dif"

    @property
    def units(self) -> str:
        """
        str : Cobalt diffusivity unit
        """
        return "[m^2.s^-1]"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Cobalt diffusivity
        correlation function
        """
        return [1023.0, 1273.0]

    @property
    def long_name(self) -> str:
        """
        str : Cobalt diffusivity long name
        """
        return "cobalt diffusivity"

    @property
    def description(self) -> str:
        """
        str : Cobalt diffusivity description
        """
        return f"{self.long_name} in liquid lead"


class SeleniumDiffusivity(PropertyInterface):
    """
    Liquid lead *Selenium diffusivity* property class.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Selenium diffusivity* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            diffusivity in :math:`[m^2 / s]`
        """
        return np.exp(-12958 / R / T) * 3.4e-8

    @property
    def name(self) -> str:
        """
        str : Name of the property
        """
        return "se_dif"

    @property
    def units(self) -> str:
        """
        str : Selenium diffusivity unit
        """
        return "[m^2.s^-1]"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Selenium diffusivity
        correlation function
        """
        return [823.0, 1173.0]

    @property
    def long_name(self) -> str:
        """
        str : Selenium diffusivity long name
        """
        return "selenium diffusivity"

    @property
    def description(self) -> str:
        """
        str : Selenium diffusivity description
        """
        return f"{self.long_name} in liquid lead"


class IndiumDiffusivity(PropertyInterface):
    """
    Liquid lead *Indium diffusivity* property class.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Indium diffusivity* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            diffusivity in :math:`[m^2 / s]`
        """
        return np.exp(-13794 / R / T) * 3.1e-8

    @property
    def name(self) -> str:
        """
        str : Name of the property
        """
        return "in_dif"

    @property
    def units(self) -> str:
        """
        str : Indium diffusivity unit
        """
        return "[m^2.s^-1]"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Indium diffusivity
        correlation function
        """
        return [723.0, 1173.0]

    @property
    def long_name(self) -> str:
        """
        str : Indium diffusivity long name
        """
        return "indium diffusivity"

    @property
    def description(self) -> str:
        """
        str : Indium diffusivity description
        """
        return f"{self.long_name} in liquid lead"


class TelluriumDiffusivity(PropertyInterface):
    """
    Liquid lead *Tellurium diffusivity* property class.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Tellurium diffusivity* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            diffusivity in :math:`[m^2 / s]`
        """
        return np.exp(-15884 / R / T) * 3.1e-8

    @property
    def name(self) -> str:
        """
        str : Name of the property
        """
        return "te_dif"

    @property
    def units(self) -> str:
        """
        str : Tellurium diffusivity unit
        """
        return "[m^2.s^-1]"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Tellurium diffusivity
        correlation function
        """
        return [723.0, 1173.0]

    @property
    def long_name(self) -> str:
        """
        str : Tellurium diffusivity long name
        """
        return "tellurium diffusivity"

    @property
    def description(self) -> str:
        """
        str : Tellurium diffusivity description
        """
        return f"{self.long_name} in liquid lead"



"""Module with the definition of the *Oxygen concentration lower limits*
objects for *lead*."""
from typing import List
from typing import Union
import numpy as np
from scipy.constants import atm
from scipy.constants import R
from ..interface import PropertyInterface
from ..tch_common_interface import LowerLimitSaturationIronInterface
from ..tch_common_interface import LowerLimitSaturationChromiumInterface
from ..tch_common_interface import LowerLimitSaturationNickelInterface
from ..tch_common_interface import LowerLimitSaturationSiliconInterface
from ..tch_common_interface import LowerLimitSaturationAluminiumInterface
from ..tch_common_interface import LowerLimitChromiumInterface
from ..tch_common_interface import LowerLimitNickelInterface
from ..tch_common_interface import LowerLimitIronInterface
from .solubility_in_lead import OxygenSolubility
from .solubility_in_lead import ChromiumSolubilityGosse2014
from .solubility_in_lead import ChromiumSolubilityAlden1958
from .solubility_in_lead import ChromiumSolubilityVenkatraman1988
from .solubility_in_lead import NickelSolubility
from .solubility_in_lead import IronSolubility
from .solubility_in_lead import SiliconSolubility
from ..._decorators import range_warning


class LowerLimitSaturationIron(LowerLimitSaturationIronInterface):
    """
    *Lower limit of Oxygen concentration* to promote a
    protective oxide film in liquid lead considering
    *Iron at its saturation concentration* property class.
    """
    def initialization_helper(self,
                              property_value: float) -> Union[None, float]:
        """
        Returns the temperature guess value according to the value
        of the lower limit of Oxygen concentration to promote a protective
        oxide film in liquid lead considering Iron at its saturation
        concentration passed as argument.
        It is used by the root finder algorithm.

        Parameters
        ----------
        property_value : float
            lower limit of Oxygen concentration to promote a
            protective oxide film in liquid lead considering
            Iron at its saturation concentration
            in :math:`[wt.\\%]`

        Returns
        -------
        float
            Temperature guess value in :math:`[K]`
        """
        if property_value < 1e-4:
            return 1000
        return 1700

    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen concentration lower
        limit* to promote a protective oxide film in liquid lead
        considering *Iron at its saturation concentration* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            Oxygen concentration in :math:`[wt.\\%]`
        """
        return np.exp(-57190 / R / T - 21.1 / R)\
            * OxygenSolubility().correlation(T, p)

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen concentration
        lower limit correlation function
        """
        return [673, 1000]

    @property
    def description(self) -> str:
        """
        str : Oxygen concentration lower limit description
        """
        return f"{self.long_name} in liquid lead"


class LowerLimitSaturationChromium(LowerLimitSaturationChromiumInterface):
    """
    *Lower limit of Oxygen concentration* to promote a
    protective oxide film in liquid lead considering
    *Chromium at its saturation concentration* property class.
    """
    def initialization_helper(self,
                              property_value: float) -> Union[None, float]:
        """
        Returns the temperature guess value according to the value
        of the lower limit of Oxygen concentration to promote a protective
        oxide film in liquid lead considering Chromium at its saturation
        concentration passed as argument.
        It is used by the root finder algorithm.

        Parameters
        ----------
        property_value : float
            lower limit of Oxygen concentration to promote a
            protective oxide film in liquid lead considering
            Chromium at its saturation concentration
            in :math:`[wt.\\%]`

        Returns
        -------
        float
            Temperature guess value in :math:`[K]`
        """
        if property_value < 1e-7:
            return 1200
        return 1700

    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen concentration lower
        limit* to promote a protective oxide film in liquid lead
        considering *Chromium at its saturation concentration* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            Oxygen concentration in :math:`[wt.\\%]`
        """
        return np.exp(-158900 / R / T - 13.65 / R)\
            * OxygenSolubility().correlation(T, p)

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen concentration
        lower limit correlation function
        """
        return [673, 1000]

    @property
    def description(self) -> str:
        """
        str : Oxygen concentration lower limit description
        """
        return f"{self.long_name} in liquid lead"


class LowerLimitSaturationNickel(LowerLimitSaturationNickelInterface):
    """
    *Lower limit of Oxygen concentration* to promote a
    protective oxide film in liquid lead considering
    *Nickel at its saturation concentration* property class.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen concentration lower
        limit* to promote a protective oxide film in liquid lead
        considering *Nickel at its saturation concentration* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            Oxygen concentration in :math:`[wt.\\%]`
        """
        return np.exp(-18040 / R / T - 11.7 / R)\
            * OxygenSolubility().correlation(T, p)

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen concentration
        lower limit correlation function
        """
        return [673, 1000]

    @property
    def description(self) -> str:
        """
        str : Oxygen concentration lower limit description
        """
        return f"{self.long_name} in liquid lead"


class LowerLimitSaturationSilicon(LowerLimitSaturationSiliconInterface):
    """
    *Lower limit of Oxygen concentration* to promote a
    protective oxide film in liquid lead considering
    *Silicon at its saturation concentration* property class.
    """
    def initialization_helper(self,
                              property_value: float) -> Union[None, float]:
        """
        Returns the temperature guess value according to the value
        of the lower limit of Oxygen concentration to promote a protective
        oxide film in liquid lead considering Silicon at its saturation
        concentration passed as argument.
        It is used by the root finder algorithm.

        Parameters
        ----------
        property_value : float
            lower limit of Oxygen concentration to promote a
            protective oxide film in liquid lead considering
            Silicon at its saturation concentration
            in :math:`[wt.\\%]`

        Returns
        -------
        float
            Temperature guess value in :math:`[K]`
        """
        if property_value < 1e-8:
            return 1400
        return 1800

    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen concentration lower
        limit* to promote a protective oxide film in liquid lead
        considering *Silicon at its saturation concentration* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            Oxygen concentration in :math:`[wt.\\%]`
        """
        return np.exp(-235855 / R / T - 9.75 / R)\
            * OxygenSolubility().correlation(T, p)

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen concentration
        lower limit correlation function
        """
        return [673, 1000]

    @property
    def description(self) -> str:
        """
        str : Oxygen concentration lower limit description
        """
        return f"{self.long_name} in liquid lead"


class LowerLimitSaturationAluminium(LowerLimitSaturationAluminiumInterface):
    """
    *Lower limit of Oxygen concentration* to promote a
    protective oxide film in liquid lead considering
    *Aluminium at its saturation concentration* property class.
    """
    def initialization_helper(self,
                              property_value: float) -> Union[None, float]:
        """
        Returns the temperature guess value according to the value
        of the lower limit of Oxygen concentration to promote a protective
        oxide film in liquid lead considering Aluminium at its saturation
        concentration passed as argument.
        It is used by the root finder algorithm.

        Parameters
        ----------
        property_value : float
            lower limit of Oxygen concentration to promote a
            protective oxide film in liquid lead considering
            Aluminium at its saturation concentration
            in :math:`[wt.\\%]`

        Returns
        -------
        float
            Temperature guess value in :math:`[K]`
        """
        return 1800

    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen concentration lower
        limit* to promote a protective oxide film in liquid lead
        considering *Aluminium at its saturation concentration* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            Oxygen concentration in :math:`[wt.\\%]`
        """
        return np.exp(-339770 / R / T + 5.35 / R)\
            * OxygenSolubility().correlation(T, p)

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen concentration
        lower limit correlation function
        """
        return [673, 1000]

    @property
    def description(self) -> str:
        """
        str : Oxygen concentration lower limit description
        """
        return f"{self.long_name} in liquid lead"


class LeadLowerLimitChromiumInterface(LowerLimitChromiumInterface):
    """
    *Lower limit of oxygen concentration* to promote a
    protective oxide film times the *Chromium concentration*
    raised to :math:`2/3` in liquid lead property abstract class.
    """
    def initialization_helper(self,
                              property_value: float) -> Union[None, float]:
        """
        Returns the temperature guess value according to the value
        of the lower limit of Oxygen concentration to promote a protective
        oxide film times the Chromium concentration raised to :math:`2/3`
        in liquid lead passed as argument.
        It is used by the root finder algorithm.

        Parameters
        ----------
        property_value : float
            lower limit of Oxygen concentration to promote a
            protective oxide film times the Chromium concentration
            raised to :math:`2/3` in liquid lead
            in :math:`[wt.\\%]`

        Returns
        -------
        float
            Temperature guess value in :math:`[K]`
        """
        if property_value < 1e-6:
            return 1500
        return 1800

    @property
    def description(self) -> str:
        """
        str : Oxygen concentration lower limit description
        """
        return f"{self.long_name} in lead"


class LowerLimitChromiumGosse2014(LeadLowerLimitChromiumInterface):
    """
    *Lower limit of oxygen concentration* to promote a
    protective oxide film times the *Chromium concentration*
    raised to :math:`2/3` in liquid lead property class
    implementing the correlation by *gosse2014*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen concentration lower
        limit* to promote a protective oxide film times Chromium
        concentration raised to :math:`2/3` in liquid lead
        by applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            Oxygen concentration in :math:`[wt.\\%]`
        """
        return LowerLimitSaturationChromium().correlation(T, p)\
            * np.power(ChromiumSolubilityGosse2014().correlation(T, p), 2 / 3)

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "gosse2014"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen concentration
        lower limit correlation function
        """
        return [673, 1000]


class LowerLimitChromiumVenkatraman1988(LeadLowerLimitChromiumInterface):
    """
    *Lower limit of oxygen concentration* to promote a
    protective oxide film times the *Chromium concentration*
    raised to :math:`2/3` in liquid lead property class
    implementing the correlation by *venkatraman1988*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen concentration lower
        limit* to promote a protective oxide film times Chromium
        concentration raised to :math:`2/3` in liquid lead
        by applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            Oxygen concentration in :math:`[wt.\\%]`
        """
        return LowerLimitSaturationChromium().correlation(T, p)\
            * np.power(ChromiumSolubilityVenkatraman1988().correlation(T, p),
                       2 / 3)

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "venkatraman1988"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen concentration
        lower limit correlation function
        """
        return [673, 1000]


class LowerLimitChromiumAlden1958(LeadLowerLimitChromiumInterface):
    """
    *Lower limit of oxygen concentration* to promote a
    protective oxide film times the *Chromium concentration*
    raised to :math:`2/3` in liquid lead property class
    implementing the correlation by *alden1958*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen concentration lower
        limit* to promote a protective oxide film times Chromium
        concentration raised to :math:`2/3` in liquid lead
        by applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            Oxygen concentration in :math:`[wt.\\%]`
        """
        return LowerLimitSaturationChromium().correlation(T, p)\
            * np.power(ChromiumSolubilityAlden1958().correlation(T, p), 2 / 3)

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "alden1958"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen concentration
        lower limit correlation function
        """
        return [673, 1000]


class LowerLimitNickel(LowerLimitNickelInterface):
    """
    *Lower limit of oxygen concentration* to promote a
    protective oxide film times the *Nickel concentration*
    in liquid lead property class.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen concentration lower
        limit* to promote a protective oxide film times Nickel
        concentration in liquid lead
        by applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            Oxygen concentration in :math:`[wt.\\%]`
        """
        return LowerLimitSaturationNickel().correlation(T, p)\
            * NickelSolubility().correlation(T, p)

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen concentration
        lower limit correlation function
        """
        return [673, 917]

    @property
    def description(self) -> str:
        """
        str : Oxygen concentration lower limit description
        """
        return f"{self.long_name} in lead"


class LowerLimitIron(LowerLimitIronInterface):
    """
    *Lower limit of oxygen concentration* to promote a
    protective oxide film times the *Iron concentration*
    raised to :math:`3/4` in liquid lead property class.
    """
    def initialization_helper(self,
                              property_value: float) -> Union[None, float]:
        """
        Returns the temperature guess value according to the value
        of the lower limit of Oxygen concentration to promote a protective
        oxide film times the Iron concentration raised to :math:`3/4`
        in liquid lead passed as argument.
        It is used by the root finder algorithm.

        Parameters
        ----------
        property_value : float
            lower limit of Oxygen concentration to promote a
            protective oxide film times the Iron concentration
            raised to :math:`3/4` in liquid lead
            in :math:`[wt.\\%]`

        Returns
        -------
        float
            Temperature guess value in :math:`[K]`
        """
        if property_value < 1e-4:
            return 1300
        return 1700

    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen concentration lower
        limit* to promote a protective oxide film times Iron
        concentration raised to :math:`3/4` in liquid lead
        by applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            Oxygen concentration in :math:`[wt.\\%]`
        """
        return LowerLimitSaturationIron().correlation(T, p)\
            * np.power(IronSolubility().correlation(T, p), 0.75)

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen concentration
        lower limit correlation function
        """
        return [673, 1000]

    @property
    def description(self) -> str:
        """
        str : Oxygen concentration lower limit description
        """
        return f"{self.long_name} in lead"


class LowerLimitSilicon(PropertyInterface):
    """
    *Lower limit of oxygen concentration* to promote a
    protective oxide film times the *Silicon concentration*
    raised to :math:`1/2` in liquid lead property class.
    """
    def initialization_helper(self,
                              property_value: float) -> Union[None, float]:
        """
        Returns the temperature guess value according to the value
        of the lower limit of Oxygen concentration to promote a protective
        oxide film times the Silicon concentration raised to :math:`1/2`
        in liquid lead passed as argument.
        It is used by the root finder algorithm.

        Parameters
        ----------
        property_value : float
            lower limit of Oxygen concentration to promote a
            protective oxide film times the Silicon concentration
            raised to :math:`1/2` in liquid lead
            in :math:`[wt.\\%]`

        Returns
        -------
        float
            Temperature guess value in :math:`[K]`
        """
        if property_value < 1e-8:
            return 1500
        return 1800

    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen concentration lower
        limit* to promote a protective oxide film times Silicon
        concentration raised to :math:`1/2` in liquid lead
        by applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            Oxygen concentration in :math:`[wt.\\%]`
        """
        return LowerLimitSaturationSilicon().correlation(T, p)\
            * np.power(SiliconSolubility().correlation(T, p), 0.5)

    @property
    def name(self) -> str:
        """
        str : Name of the property
        """
        return "lim_si"

    @property
    def units(self) -> str:
        """
        str : Oxygen concentration lower limit unit
        """
        return "[wt.%]"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen concentration
        lower limit correlation function
        """
        return [673, 1000]

    @property
    def long_name(self) -> str:
        """
        str : Oxygen concentration lower limit long name
        """
        return ("Oxygen concentration lower limit times"
                " silicon concentration raised to 1/2")

    @property
    def description(self) -> str:
        """
        str : Oxygen concentration lower limit description
        """
        return f"{self.long_name} in lead"



"""Module with the definition of the *Oxygen concentration lower limits*
objects for *lead-bismuth eutectic* (*lbe*)."""
from typing import List
from typing import Union
import numpy as np
from scipy.constants import atm
from scipy.constants import R
from ..tch_common_interface import LowerLimitSaturationIronInterface
from ..tch_common_interface import LowerLimitSaturationChromiumInterface
from ..tch_common_interface import LowerLimitSaturationNickelInterface
from ..tch_common_interface import LowerLimitSaturationSiliconInterface
from ..tch_common_interface import LowerLimitSaturationAluminiumInterface
from ..tch_common_interface import LowerLimitChromiumInterface
from ..tch_common_interface import LowerLimitNickelInterface
from ..tch_common_interface import LowerLimitIronInterface
from .solubility_in_lbe import OxygenSolubility
from .solubility_in_lbe import ChromiumSolubilityGosse2014
from .solubility_in_lbe import ChromiumSolubilityCourouau2004
from .solubility_in_lbe import ChromiumSolubilityMartynov1998
from .solubility_in_lbe import NickelSolubilityGosse2014
from .solubility_in_lbe import NickelSolubilityMartinelli2010
from .solubility_in_lbe import IronSolubilityGosse2014
from .solubility_in_lbe import IronSolubilityWeeks1969
from .lbe_thermochemical import LeadChemicalActivity
from ..._decorators import range_warning


class LowerLimitSaturationIron(LowerLimitSaturationIronInterface):
    """
    *Lower limit of Oxygen concentration* to promote a
    protective oxide film in liquid lbe considering
    *Iron at its saturation concentration* property class.
    """
    def initialization_helper(self,
                              property_value: float) -> Union[None, float]:
        """
        Returns the temperature guess value according to the value
        of the lower limit of Oxygen concentration to promote a protective
        oxide film in liquid lbe considering Iron at its saturation
        concentration passed as argument.
        It is used by the root finder algorithm.

        Parameters
        ----------
        property_value : float
            lower limit of Oxygen concentration to promote a
            protective oxide film in liquid lbe considering
            Iron at its saturation concentration
            in :math:`[wt.\\%]`

        Returns
        -------
        float
            Temperature guess value in :math:`[K]`
        """
        return 1700

    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen concentration lower
        limit* to promote a protective oxide film in liquid lbe
        considering *Iron at its saturation concentration* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            Oxygen concentration in :math:`[wt.\\%]`
        """
        return np.exp(-57190 / R / T - 21.1 / R)\
            * OxygenSolubility().correlation(T, p)\
            * LeadChemicalActivity().correlation(T, p)

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen concentration
        lower limit correlation function
        """
        return [673, 1000]

    @property
    def description(self) -> str:
        """
        str : Oxygen concentration lower limit description
        """
        return f"{self.long_name} in liquid lbe"


class LowerLimitSaturationChromium(LowerLimitSaturationChromiumInterface):
    """
    *Lower limit of Oxygen concentration* to promote a
    protective oxide film in liquid lbe considering
    *Chromium at its saturation concentration* property class.
    """
    def initialization_helper(self,
                              property_value: float) -> Union[None, float]:
        """
        Returns the temperature guess value according to the value
        of the lower limit of Oxygen concentration to promote a protective
        oxide film in liquid lbe considering Chromium at its saturation
        concentration passed as argument.
        It is used by the root finder algorithm.

        Parameters
        ----------
        property_value : float
            lower limit of Oxygen concentration to promote a
            protective oxide film in liquid lbe considering
            Chromium at its saturation concentration
            in :math:`[wt.\\%]`

        Returns
        -------
        float
            Temperature guess value in :math:`[K]`
        """
        return 1700

    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen concentration lower
        limit* to promote a protective oxide film in liquid lbe
        considering *Chromium at its saturation concentration* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            Oxygen concentration in :math:`[wt.\\%]`
        """
        return np.exp(-158900 / R / T - 13.65 / R)\
            * OxygenSolubility().correlation(T, p)\
            * LeadChemicalActivity().correlation(T, p)

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen concentration
        lower limit correlation function
        """
        return [673, 1000]

    @property
    def description(self) -> str:
        """
        str : Oxygen concentration lower limit description
        """
        return f"{self.long_name} in lbe"


class LowerLimitSaturationNickel(LowerLimitSaturationNickelInterface):
    """
    *Lower limit of Oxygen concentration* to promote a
    protective oxide film in liquid lbe considering
    *Nickel at its saturation concentration* property class.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen concentration lower
        limit* to promote a protective oxide film in liquid lbe
        considering *Nickel at its saturation concentration* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            Oxygen concentration in :math:`[wt.\\%]`
        """
        return np.exp(-18040 / R / T - 11.7 / R)\
            * OxygenSolubility().correlation(T, p)\
            * LeadChemicalActivity().correlation(T, p)

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen concentration
        lower limit correlation function
        """
        return [673, 1000]

    @property
    def description(self) -> str:
        """
        str : Oxygen concentration lower limit description
        """
        return f"{self.long_name} in liquid lbe"


class LowerLimitSaturationSilicon(LowerLimitSaturationSiliconInterface):
    """
    *Lower limit of Oxygen concentration* to promote a
    protective oxide film in liquid lbe considering
    *Silicon at its saturation concentration* property class.
    """
    def initialization_helper(self,
                              property_value: float) -> Union[None, float]:
        """
        Returns the temperature guess value according to the value
        of the lower limit of Oxygen concentration to promote a protective
        oxide film in liquid lbe considering Silicon at its saturation
        concentration passed as argument.
        It is used by the root finder algorithm.

        Parameters
        ----------
        property_value : float
            lower limit of Oxygen concentration to promote a
            protective oxide film in liquid lbe considering
            Silicon at its saturation concentration
            in :math:`[wt.\\%]`

        Returns
        -------
        float
            Temperature guess value in :math:`[K]`
        """
        if property_value < 1e-9:
            return 1400
        return 1700

    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen concentration lower
        limit* to promote a protective oxide film in liquid lbe
        considering *Silicon at its saturation concentration* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            Oxygen concentration in :math:`[wt.\\%]`
        """
        return np.exp(-235855 / R / T - 9.75 / R)\
            * OxygenSolubility().correlation(T, p)\
            * LeadChemicalActivity().correlation(T, p)

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen concentration
        lower limit correlation function
        """
        return [673, 1000]

    @property
    def description(self) -> str:
        """
        str : Oxygen concentration lower limit description
        """
        return f"{self.long_name} in liquid lbe"


class LowerLimitSaturationAluminium(LowerLimitSaturationAluminiumInterface):
    """
    *Lower limit of Oxygen concentration* to promote a
    protective oxide film in liquid lbe considering
    *Aluminium at its saturation concentration* property class.
    """
    def initialization_helper(self,
                              property_value: float) -> Union[None, float]:
        """
        Returns the temperature guess value according to the value
        of the lower limit of Oxygen concentration to promote a protective
        oxide film in liquid lbe considering Aluminium at its saturation
        concentration passed as argument.
        It is used by the root finder algorithm.

        Parameters
        ----------
        property_value : float
            lower limit of Oxygen concentration to promote a
            protective oxide film in liquid lbe considering
            Aluminium at its saturation concentration
            in :math:`[wt.\\%]`

        Returns
        -------
        float
            Temperature guess value in :math:`[K]`
        """
        if property_value < 1e-11:
            return 1500
        return 1800

    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen concentration lower
        limit* to promote a protective oxide film in liquid lbe
        considering *Aluminium at its saturation concentration* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            Oxygen concentration in :math:`[wt.\\%]`
        """
        return np.exp(-339770 / R / T + 5.35 / R)\
            * OxygenSolubility().correlation(T, p)\
            * LeadChemicalActivity().correlation(T, p)

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen concentration
        lower limit correlation function
        """
        return [673, 1000]

    @property
    def description(self) -> str:
        """
        str : Oxygen concentration lower limit description
        """
        return f"{self.long_name} in liquid lbe"


class LBELowerLimitChromiumInterface(LowerLimitChromiumInterface):
    """
    *Lower limit of oxygen concentration* to promote a
    protective oxide film times the *Chromium concentration*
    raised to :math:`2/3` in liquid lbe property abstract class.
    """
    def initialization_helper(self,
                              property_value: float) -> Union[None, float]:
        """
        Returns the temperature guess value according to the value
        of the lower limit of Oxygen concentration to promote a protective
        oxide film times the Chromium concentration raised to :math:`2/3`
        in liquid lbe passed as argument.
        It is used by the root finder algorithm.

        Parameters
        ----------
        property_value : float
            lower limit of Oxygen concentration to promote a
            protective oxide film times the Chromium concentration
            raised to :math:`2/3` in liquid lbe
            in :math:`[wt.\\%]`

        Returns
        -------
        float
            Temperature guess value in :math:`[K]`
        """
        if property_value < 1e-7:
            return 1400
        return 1700

    @property
    def description(self) -> str:
        """
        str : Oxygen concentration lower limit description
        """
        return f"{self.long_name} in lbe"


class LowerLimitChromiumGosse2014(LBELowerLimitChromiumInterface):
    """
    *Lower limit of oxygen concentration* to promote a
    protective oxide film times the *Chromium concentration*
    raised to :math:`2/3` in liquid lbe property class
    implementing the correlation by *gosse2014*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen concentration lower
        limit* to promote a protective oxide film times Chromium
        concentration raised to :math:`2/3` in liquid lbe
        by applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            Oxygen concentration in :math:`[wt.\\%]`
        """
        return LowerLimitSaturationChromium().correlation(T, p)\
            * np.power(ChromiumSolubilityGosse2014().correlation(T, p), 2 / 3)

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "gosse2014"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen concentration
        lower limit correlation function
        """
        return [673, 1000]


class LowerLimitChromiumCourouau2004(LBELowerLimitChromiumInterface):
    """
    *Lower limit of oxygen concentration* to promote a
    protective oxide film times the *Chromium concentration*
    raised to :math:`2/3` in liquid lbe property class
    implementing the correlation by *courouau2004*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen concentration lower
        limit* to promote a protective oxide film times Chromium
        concentration raised to :math:`2/3` in liquid lbe
        by applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            Oxygen concentration in :math:`[wt.\\%]`
        """
        return LowerLimitSaturationChromium().correlation(T, p)\
            * np.power(ChromiumSolubilityCourouau2004().correlation(T, p),
                       2 / 3)

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "courouau2004"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen concentration
        lower limit correlation function
        """
        return [673, 813]


class LowerLimitChromiumMartynov1998(LBELowerLimitChromiumInterface):
    """
    *Lower limit of oxygen concentration* to promote a
    protective oxide film times the *Chromium concentration*
    raised to :math:`2/3` in liquid lbe property class
    implementing the correlation by *martynov1998*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen concentration lower
        limit* to promote a protective oxide film times Chromium
        concentration raised to :math:`2/3` in liquid lbe
        by applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            Oxygen concentration in :math:`[wt.\\%]`
        """
        return LowerLimitSaturationChromium().correlation(T, p)\
            * np.power(ChromiumSolubilityMartynov1998().correlation(T, p),
                       2 / 3)

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "martynov1998"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen concentration
        lower limit correlation function
        """
        return [673, 773]


class LBELowerLimitNickelInterface(LowerLimitNickelInterface):
    """
    *Lower limit of oxygen concentration* to promote a
    protective oxide film times the *Nickel concentration*
    in liquid lbe property abstract class.
    """
    @property
    def description(self) -> str:
        """
        str : Oxygen concentration lower limit description
        """
        return f"{self.long_name} in lbe"


class LowerLimitNickelMartinelli2010(LBELowerLimitNickelInterface):
    """
    *Lower limit of oxygen concentration* to promote a
    protective oxide film times the *Nickel concentration*
    in liquid lbe property class
    implementing the correlation by *martinelli2010*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> Union[float, np.ndarray]:
        """
        Returns the value of the *Oxygen concentration lower
        limit* to promote a protective oxide film times Nickel
        concentration in liquid lbe
        by applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            Oxygen concentration in :math:`[wt.\\%]`
        """
        return LowerLimitSaturationNickel().correlation(T, p)\
            * NickelSolubilityMartinelli2010().correlation(T, p)

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "martinelli2010"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen concentration
        lower limit correlation function
        """
        return [673, 1000]


class LowerLimitNickelGosse2014(LBELowerLimitNickelInterface):
    """
    *Lower limit of oxygen concentration* to promote a
    protective oxide film times the *Nickel concentration*
    in liquid lbe property class
    implementing the correlation by *gosse2014*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> Union[float, np.ndarray]:
        """
        Returns the value of the *Oxygen concentration lower
        limit* to promote a protective oxide film times Nickel
        concentration in liquid lbe
        by applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            Oxygen concentration in :math:`[wt.\\%]`
        """
        return LowerLimitSaturationNickel().correlation(T, p)\
            * NickelSolubilityGosse2014().correlation(T, p)

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "gosse2014"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen concentration
        lower limit correlation function
        """
        return [673, 1000]


class LBELowerLimitIronInterface(LowerLimitIronInterface):
    """
    *Lower limit of oxygen concentration* to promote a
    protective oxide film times the *Iron concentration*
    raised to :math:`3/4` in liquid lbe property abstract class.
    """
    def initialization_helper(self,
                              property_value: float) -> Union[None, float]:
        """
        Returns the temperature guess value according to the value
        of the lower limit of Oxygen concentration to promote a protective
        oxide film times the Iron concentration raised to :math:`3/4`
        in liquid lbe passed as argument.
        It is used by the root finder algorithm.

        Parameters
        ----------
        property_value : float
            lower limit of Oxygen concentration to promote a
            protective oxide film times the Iron concentration
            raised to :math:`3/4` in liquid lbe
            in :math:`[wt.\\%]`

        Returns
        -------
        float
            Temperature guess value in :math:`[K]`
        """
        if property_value < 1e-6:
            return 1200
        return 1700

    @property
    def description(self) -> str:
        """
        str : Oxygen concentration lower limit description
        """
        return f"{self.long_name} in lbe"


class LowerLimitIronGosse2014(LBELowerLimitIronInterface):
    """
    *Lower limit of oxygen concentration* to promote a
    protective oxide film times the *Iron concentration*
    raised to :math:`3/4` in liquid lbe property class
    implementing the correlation by *gosse2014*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen concentration lower
        limit* to promote a protective oxide film times Iron
        concentration raised to :math:`3/4` in liquid lbe
        by applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            Oxygen concentration in :math:`[wt.\\%]`
        """
        return LowerLimitSaturationIron().correlation(T, p)\
            * np.power(IronSolubilityGosse2014().correlation(T, p), 0.75)

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "gosse2014"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen concentration
        lower limit correlation function
        """
        return [673, 1000]


class LowerLimitIronWeeks1969(LBELowerLimitIronInterface):
    """
    *Lower limit of oxygen concentration* to promote a
    protective oxide film times the *Iron concentration*
    raised to :math:`3/4` in liquid lbe property class
    implementing the correlation by *weeks1969*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen concentration lower
        limit* to promote a protective oxide film times Iron
        concentration raised to :math:`3/4` in liquid lbe
        by applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            Oxygen concentration in :math:`[wt.\\%]`
        """
        return LowerLimitSaturationIron().correlation(T, p)\
            * np.power(IronSolubilityWeeks1969().correlation(T, p), 0.75)

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "weeks1969"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen concentration
        lower limit correlation function
        """
        return [673, 1000]



"""Module with the definition of liquid metal object base class,
i.e., LiquidMetalInterface"""
import warnings
import sys
import inspect
import importlib
import pathlib
import copy
from abc import ABC
from abc import abstractmethod
from collections import defaultdict
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union
from scipy.constants import atm
from scipy.optimize import fsolve
from .properties.interface import PropertyInterface

warnings.simplefilter("always")


class LiquidMetalInterface(ABC):
    """
    Abstract class that defines liquid metal properties object

    Parameters
    ----------
    p : float, optional
        Pressure in [Pa], by default the atmospheric pressure value, i.e.,
        101325.0 Pa
    **kwargs : dict
        One-item dictionary that specifies the quantity which the object shall
        be initialized from. The default available ones are:

        - 'T' : temperature [K]
        - 'p_s' : saturation vapour pressure [Pa]
        - 'sigma' : surface tension [N/m]
        - 'rho' : density [Kg/m^3]
        - 'alpha' : thermal expansion coefficient [1/K]
        - 'u_s': speed of sound [m/s]
        - 'beta_s' : isentropic compressibility [1/Pa]
        - 'cp' : specific heat capacity [J/(kg \\cdot K)]
        - 'h' : specific enthalpy (in respect to melting point) [J/kg]
        - 'mu' : dynamic viscosity [Pa*s]
        - 'r' : electrical resistivity [Ohm*m]
        - 'k' : thermal conductivity [W/(m*K)]
    """
    _T_m0: float = 0
    _Q_m0: float = 0
    _T_b0: float = 0
    _Q_b0: float = 0
    _M: float = 0
    _guess: float = 0
    _correlations_to_use: Dict[str, str] = {}
    _roots_to_use: Dict[str, int] = {}
    _default_corr_to_use: Dict[str, str] = {}
    _properties_modules_list: List[str] = []
    _custom_properties_path: Dict[str, List[str]] = {}
    _available_properties_dict: Dict[str, PropertyInterface] = {}
    _available_correlations_dict: Dict[str, List[str]] = {}
    __p: float = 0
    __T: float = 0

    def __init__(self, p: float = atm, **kwargs):
        if len(kwargs) != 1:
            raise ValueError("One and only one property at "
                             "time can be used for initialization. "
                             f"{len(kwargs)} were provided")
        self.__assign_p(p)
        self.__properties: Dict[str, PropertyInterface] = {}
        self.__corr2use: Dict[str, str] = \
            copy.deepcopy(self.__class__._correlations_to_use)
        self.__fill_instance_properties()
        self._set_constants()
        name, value = kwargs.popitem()
        self.__fill_instance_attributes(name, value)

    @property
    def T_m0(self) -> float:
        """
        float : melting temperature :math:`[K]`
        """
        return self._T_m0

    @property
    def Q_m0(self) -> float:
        """
        float : melting latent heat :math:`[J/kg]`
        """
        return self._Q_m0

    @property
    def T_b0(self) -> float:
        """
        float : boiling temperature :math:`[K]`
        """
        return self._T_b0

    @property
    def Q_b0(self) -> float:
        """
        float : vaporisation heat :math:`[J/kg]`
        """
        return self._Q_b0

    @property
    def M(self) -> float:
        """
        float : metal molar mass :math:`[g/mol]`
        """
        return self._M

    @property
    def p(self) -> float:
        """
        float : pressure adopted for property calculation :math:`[Pa]`, \
            i.e., atmospheric pressure
        """
        return self.__p

    @p.setter
    def p(self, p: float) -> None:
        """
        p(p: float) -> None

        Set liquid metal pressure

        Parameters
        ----------
        p : float
            Pressure in [Pa]
        """
        self.__assign_p(p)

    @property
    def T(self) -> float:
        """
        float : temperature used to compute properties :math:`[K]`
        """
        return self.__T

    @T.setter
    def T(self, T: float) -> None:
        """
        Set liquid metal temperature

        Parameters
        ----------
        T : float
            Temperature in [K]
        """
        self.__assign_T(T)

    @property
    def Pr(self) -> float:
        """
        float : Prandtl number :math:`[-]`
        """
        return self.cp * self.mu / self.k

    @property
    def used_correlations(self) -> Dict[str, str]:
        """
        Returns the dictionary defining the correlation used for the \
        corresponding property. Only properties are considered for \
        which a specific correlation has been chosen a priori.

        Returns
        -------
        Dict[str, str]
        """
        return copy.deepcopy(self.__corr2use)

    def change_correlation_to_use(self, property_name: str,
                                  correlation_name: str) -> None:
        """
        Changes property correlation if property is defined as instance
        attribute, otherwise a warning message is returned without
        applying any change.

        Parameters
        ----------
        property_name : str
            Name of the property
        correlation_name : str
            Name of the correlation
        """
        # Manage the case the property is not among the currently used ones
        if property_name not in self.__properties:
            warnings.warn(f"'{property_name}' property not in use."
                          "\nNothing to change.", stacklevel=5)
            return
        # Manage the case the input correlation is already used
        if (self.__properties[property_name].correlation_name
                == correlation_name):
            warnings.warn(f"'{property_name}' property "
                          f"implementing '{correlation_name}' correlation "
                          "already in use. \nNothing to change.",
                          stacklevel=5)
            return
        # Manage the case the input correlation not among the available ones
        key = property_name + "__" + correlation_name
        if key not in self._available_properties_dict:
            warnings.warn(f"'{property_name}' property implementing "
                          f"'{correlation_name}' correlation not among"
                          "the available ones. \nNothing to change.",
                          stacklevel=5)
            return
        # If here, the input correlation is to apply
        self.__add_property(self._available_properties_dict[key])
        if property_name in self._default_corr_to_use:
            self.__corr2use[property_name] = correlation_name

    def check_temperature(self, T: float) -> Tuple[bool, str]:
        """
        Checks whether the provided temperature value belongs to the valid \
            temperature range, i.e., the liquid temperature range.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`

        Returns
        -------
        Tuple[bool, str]:
            bool:
                `True` if check is ok, `False` otherwise
            str:
                error message (if any) associated
                to the temperature check
        """
        # Manage acceptable value
        if self.T_m0 < T < self.T_b0:
            return True, ""
        # Manage value outside the acceptable range
        if T >= self.T_b0:
            error_message = ("Temperature must be smaller than "
                             f"boiling temperature ({self.T_b0:.2f} [K]), "
                             f"{T:.2f} [K] was provided")
        elif 0 < T <= self.T_m0:
            error_message = ("Temperature must be larger than "
                             f"melting temperature ({self.T_m0:.2f} [K]), "
                             f"{T:.2f} [K] was provided")
        else:
            error_message = ("Temperature must be strictly positive, "
                             f"{T:.2f} [K] was provided")
        return False, error_message

    @classmethod
    def properties_for_initialization(cls) -> List[str]:
        """
        Returns the list of the available properties an instance can be
        initialized from.

        Returns
        -------
        List[str]
        """
        obj_list = cls.__load_properties()
        obj_list += cls.__load_custom_properties()
        rvalue = ['T'] + [obj_list[i].name for i in range(len(obj_list))]
        return list(dict.fromkeys(rvalue))

    @classmethod
    def available_correlations(cls,
                               properties: Union[str, List[str], None] = None)\
            -> Dict[str, List[str]]:
        """
        Returns the available correlations for the properties passed as
        argument. Result is formatted as dictionary, where keys are the
        required property names and values are the corresponding lists
        of available correlations. In case at least one required property
        is not among the implemented ones, a warning message is returned
        listing the names of all the properties that have not been found.
        In case no value is passed as argument, the available correlations
        are returned for all the implemented properties.

        Parameters
        ----------
        properties : str | List[str] | None
            name(s) of the property(ies) whose available correlations
            are to be retrieved. If multiple properties are required,
            the list of their names must be provided, otherwise a simple
            string is enough. If the correlations are required for all the
            implemented properties, then `None` value is to be passed.

        Returns
        -------
        Dict[str, List[str]]
        """
        obj_list = cls.__load_properties()
        obj_list += cls.__load_custom_properties()
        props_dict = cls.__extract_available_correlations(obj_list)

        # If no argument is passed as input,
        # return correlations for all the implemented properties
        if properties is None:
            return props_dict

        # Check for any required property that is not available
        if isinstance(properties, str):
            properties = [properties]
        props_not_avail = [
            pr for pr in properties if pr not in props_dict.keys()]
        if len(props_not_avail) > 0:
            warnings.warn(f"Required '{props_not_avail}' properties not found!"
                          "\nPlease check the property name(s) "
                          "and try again!", stacklevel=5)

        return {k: v for k, v in props_dict.items() if k in properties}

    @classmethod
    def set_correlation_to_use(cls, property_name: str,
                               correlation_name: str) -> None:
        """
        Sets the correlation to use for the specified property.

        Parameters
        ----------
        property_name : str
            Name of the property
        correlation_name : str
            Name of the correlation
        """
        cls._correlations_to_use[property_name] = correlation_name

    @classmethod
    def set_root_to_use(cls, property_name: str, root_index: int) -> None:
        """
        Sets the index of the root to use for computing the temperature \
        from the correlation function of the property the instantiation \
        is based on. Temperature roots are sorted in ascending order, i.e, \
        :math:`T_i <= T_j`, with :math:`i < j`. Used only if property \
        correlation is not injective.

        Parameters
        ----------
        property_name : str
            Name of the property
        root_index : int
            Index of the temperature root to use
        """
        cls._roots_to_use[property_name] = root_index

    @classmethod
    def set_custom_properties_path(cls, file_path: str) -> None:
        """
        Sets the absolute path of the file where looking for \
        custom properties.

        Parameters
        ----------
        file_path : str
            absolute path of the file where custom properties are implemented
        """
        # Normalize the path, no matter the OS
        norm_path = pathlib.Path(file_path)

        # Exit if the file passed as argument does not exist
        if not norm_path.exists():
            warnings.warn(f"'{norm_path}' provided file not found!"
                          "\nPlease check the file path and try again!"
                          "\nNo custom property added.", stacklevel=5)
            return

        # Add filename and the corresponding path to the class dict
        file_name = norm_path.stem
        path = str(norm_path.parent)
        if path not in cls._custom_properties_path:
            cls._custom_properties_path[path] = [file_name]
        else:
            cls._custom_properties_path[path].append(file_name)

    @classmethod
    def correlations_to_use(cls) -> Dict[str, str]:
        """
        Returns the dictionary defining the correlation used for the \
        corresponding property. Only properties are considered for \
        which a specific correlation has been chosen a priori.

        Returns
        -------
        Dict[str, str]
        """
        return copy.deepcopy(cls._correlations_to_use)

    @classmethod
    def roots_to_use(cls) -> Dict[str, int]:
        """
        Returns the dictionary containing, for each specified property, \
        the index of the root to use for identifying the temperature \
        of the liquid metal. Useful when the initialization is \
        performed from a property whose correlation is not injective.

        Returns
        -------
        Dict[str, int]
        """
        return copy.deepcopy(cls._roots_to_use)

    def __compute_T(self, input_value: float, input_property: str) -> float:
        """
        Computes the temperature in [K] that is then set as value
        of the object property

        Parameters
        ----------
        input_value : float
            value of the property used to compute the temperature
        input_property : str
            name of the property used to perform the calculation
        """
        # Manage the simplest case
        if input_property == 'T':
            return input_value

        # Retrieve the correlation to solve
        function_of_T = self.__properties[input_property].correlation
        if function_of_T is None:
            raise UnboundLocalError("No correlation found for property "
                                    f"{input_property}! The temperature "
                                    "value can not be computed!")

        def function_to_solve(T: float, target: float) -> float:
            return function_of_T(T, self.__p) - target

        # Retrieve the initial guess, if any
        helper = self.__properties[input_property].initialization_helper
        if helper is not None and helper(input_value) is not None:
            self._guess = helper(input_value) or 0.0

        # Solve the correlation-derived function depending on the
        # injectivity of the property correlation
        if self.__properties[input_property].is_injective:
            index = 0
            res, _, ier, msg = fsolve(function_to_solve, x0=[self._guess],
                                      args=(input_value), xtol=1e-10,
                                      full_output=True)
        else:
            index = (self._roots_to_use[input_property]
                     if input_property in self._roots_to_use else 0)
            res, _, ier, msg = fsolve(function_to_solve,
                                      x0=[self._guess, 3*self._guess],
                                      args=(input_value), xtol=1e-10,
                                      full_output=True)
        # Raise an exception in case the solver did not converge
        if ier == 0:
            raise RuntimeError(f"Error: {msg}\n"
                               "The temperature value can not be computed!")
        # Select the desired root
        if len(res) > index - 1:
            return res[index]
        return res[0]

    def __fill_instance_properties(self) -> None:
        """
        Fills instance properties.
        """
        # Build the dict attribute storing all the property objects
        # loaded from loaded modules and the dict attribute storing
        # all the corresponding available correlations. Both these actions
        # are performed only once, when an instance is built, that is,
        # when the property dict attribute is empty.
        if len(self._available_properties_dict) == 0:
            available_properties_list = self.__load_properties()
            available_properties_list += self.__load_custom_properties()
            self._available_properties_dict = {e.name + '__' +
                                               e.correlation_name: e for e in
                                               available_properties_list}
            self._available_correlations_dict = \
                self.__extract_available_correlations(
                    available_properties_list)

        for key, property_object in self._available_properties_dict.items():
            name = key.split("__")[0]
            # Add the property in case the specific correlation is not
            # specified or it is specified and the correlation names
            # does not match with what already stored
            if not self.__corr2use or name not in self.__corr2use.keys()\
                    or key.split("__")[1] == self.__corr2use[name]:
                self.__add_property(property_object)

        self.__align_corrs_to_properties()

    def __fill_instance_attributes(self, property_name: str,
                                   property_value: float) -> None:
        """
        Fills all the class attributes.

        Parameters
        ----------
        property_name : str
            name of the property the liquid metal instance
            is initialized upon
        property_value: float
            value of the property the liquid metal instance
            is initialized upon
        """
        valid_prop = set(['T'] + [p.split("__")[0] for p in
                                  self._available_properties_dict])
        if property_name not in valid_prop:
            list_to_print = "\n\n"
            list_to_print += '\n'.join(valid_prop)
            list_to_print += "\n\n"
            raise ValueError("Initialization can be done only with one of "
                             f"the following properties:{list_to_print}"
                             f"{property_name} was provided")

        temperature = self.__compute_T(property_value, property_name)
        self.__assign_T(temperature)

    def __assign_T(self, T: float) -> None:
        """
        Function used to set temperature, checking if temperature value
        is valid, i.e., strictly positive and (if yes) inside
        the range of melting and boiling temperature.

        Parameters
        ----------
        T : float
            Temperature in [K]
        """
        temp_ok, error_message = self.check_temperature(T)
        if not temp_ok:
            raise ValueError(error_message)
        self.__T = T

    def __assign_p(self, p: float) -> None:
        """
        Function used to set pressure, checking if
        pressure value is valid, i.e., strictly positive.

        Parameters
        ----------
            p : float
            Pressure in [Pa]
        """
        if p <= 0:
            raise ValueError("Pressure must be strictly positive, "
                             f"{p:.2f} [Pa] was provided")
        self.__p = p

    def __add_property(self, property_object: PropertyInterface) -> None:
        """
        Adds the property to class attributes. In particular, it adds
        '<prpertyObject.name>_info' as class method to get additional
        information on property. Moreover it adds property_object
        to instance dictionary which will be used in dunder __getattr__
        to return attribute '<prpertyObject.name>'.

        Parameters
        ----------
        property_object : :class:`_properties.PropertyInterface`
            Object which inherits from :class:`_properties.PropertyInterface`
        """
        key = property_object.name
        self.__properties[key] = property_object
        setattr(self, property_object.name+"_info",
                lambda print_info=True, n_tab=0:
                    self.__properties[key].info(self.__T, self.__p,
                                                print_info, n_tab))

    def __align_corrs_to_properties(self) -> None:
        """
        This method updates the instance correlation dictionary by
        aligning it to the instance property objects dict. In particular,
        it cleans the dict containing the correlations to use and,
        accordingly, it adapts the property objects dict
        """
        # Copy the corrs dict for freezing the dict to loop over
        __corr2use_ref = copy.deepcopy(self.__corr2use)
        for key in __corr2use_ref.keys():
            corr_name = __corr2use_ref[key]
            is_in_default = key in self._default_corr_to_use

            if key not in self.__properties:
                if not is_in_default:
                    warnings.warn(f"Could not find '{key}' property "
                                  f"implementing '{corr_name}' correlation. "
                                  f"\nGoing to restore {key} property from "
                                  f"the {type(self)}-related modules, "
                                  "if any.",
                                  stacklevel=5)
                    self.__remove_property(key)
                    if key in self._available_correlations_dict:
                        self.__add_property(
                            self._available_properties_dict[
                                key + "__" +
                                self._available_correlations_dict[key][-1]])
                else:
                    def_corr_name = self._default_corr_to_use[key]
                    warnings.warn(f"Could not find property '{key}' "
                                  f"implementing '{corr_name}' correlation. "
                                  "\nGoing to restore default correlation "
                                  f"'{def_corr_name}'.",
                                  stacklevel=5)
                    self.__corr2use[key] = def_corr_name
                    self.__add_property(
                        self._available_properties_dict[
                            key + "__" + def_corr_name])
            else:
                if corr_name != self.__properties[key].correlation_name:
                    warnings.warn(f"Could not find property '{key}' "
                                  f"implementing '{corr_name}' correlation. "
                                  "\nGoing to remove it from correlations "
                                  "to use.",
                                  stacklevel=5)
                    if is_in_default:
                        self.__corr2use[key] = \
                            self.__properties[key].correlation_name
                    else:
                        self.__remove_property(key)

    @classmethod
    def __load_custom_properties(cls) -> List[PropertyInterface]:
        """
        Load custom property objects

        Returns
        -------
        list
            list of property objects, i.e. of classes which inherit from
            :class:`_properties.PropertyInterface`
        """
        customproperty_obj_list = []
        modules = []
        for path, modules_list in cls._custom_properties_path.items():
            if path not in sys.path:
                sys.path.append(path)
            for module in modules_list:
                importlib.import_module(module)
                modules.append(module)
            customproperty_obj_list += cls.__load_properties(modules)
        return customproperty_obj_list

    @classmethod
    def __load_properties(cls,
                          modules: Union[List[str], None] = None
                          ) -> List[PropertyInterface]:
        """
        Loads property objects corresponding to liquid metal. The list of
        module names can be passed as argument, otherwise the class-related
        list is adopted.

        Parameters
        ----------
        modules : :obj:`typing.List`, optional
            list of module names to read the property objects from; if
            `None`, the class-related list is adopted. By default, `None`
            is passed.

        Returns
        -------
        list
            list of property objects, i.e. of classes which inherit from
            :class:`_properties.PropertyInterface`
        """
        # Consider default modules if none is provided
        if modules is None:
            modules = cls._properties_modules_list

        def is_valid(obj):
            '''
            Predicate for filtering the modules to load
            '''
            return inspect.isclass(obj) and obj is not PropertyInterface \
                and not inspect.isabstract(obj) \
                and issubclass(obj, PropertyInterface)

        # Filter any empty module
        eff_modules = [module for module in modules if module]
        # Collect all valid properties neglecting duplicates
        mod = []
        for module in eff_modules:
            mod += inspect.getmembers(sys.modules[module], is_valid)
        mod_set = set(list(map(list, zip(*mod)))[1])
        # Build property instances and add them to the list to return
        prop_list = []
        for prop in mod_set:
            prop_obj = prop()
            # Not allowed properties whose name begins with "__"
            if not prop_obj.name[:2] == '__':
                prop_list.append(prop_obj)
        return prop_list

    @staticmethod
    def __extract_available_correlations(
            prop_obj_list: List[PropertyInterface]) -> Dict[str, List[str]]:
        """
        Private static method for extracting the available correlations
        from the list collecting all the available property classes.

        Parameters
        ----------
        prop_obj_list : :obj:`typing.List`
            list of :class:`_properties.PropertyInterface` instances
            which to extract the available correlations from

        Returns
        -------
        dict
            dictionary collecting all the property class names as keys
            together with the list collecting the corresponding
            available correlation names as values
        """
        avail_corrs = defaultdict(list)
        for prop in prop_obj_list:
            avail_corrs[prop.name].append(prop.correlation_name)
        return avail_corrs

    def __remove_property(self, property_name: str) -> None:
        """
        Private method for removing the property name passed
        as argument from both the instance and the class dicts
        storing the properties currently used together with
        their corresponding correlation.

        Parameters
        ----------
        property_name : str
            name of the property to remove from both the instance and
            the class dicts storing the properties currently used together with
            their corresponding correlation.

        Returns
        -------
        None
        """
        self.__corr2use.pop(property_name)
        self._correlations_to_use.pop(property_name, None)

    @abstractmethod
    def _set_constants(self) -> None:
        """
        Sets the class constants
        """
        raise NotImplementedError(f"{type(self).__name__}._set_constants "
                                  "NOT IMPLEMENTED")

    def __getattr__(self, name: str):
        if name[:2] == '__' or name not in self.__properties:
            raise AttributeError(f"'{type(self).__name__}' object "
                                 f"has no attribute '{name}'")

        return self.__properties[name].correlation(self.__T, self.__p, True)

    def __str__(self) -> str:
        rvalue = (f"{type(self).__name__} liquid metal "
                  f"@(T={self.T:.2f} [K], p={self.p:.2f} [Pa])\n")
        rvalue += "\nConstants:\n"
        rvalue += f"\tMelting Temperature: {self.T_m0:.2f} [K]\n"
        rvalue += f"\tBoiling Temperature: {self.T_b0:.2f} [K]\n"
        rvalue += f"\tMelting latent heat: {self.Q_m0:.2f} [J/kg]\n"
        rvalue += f"\tVaporisation heat: {self.Q_b0:.2f} [J/kg]\n"
        rvalue += f"\tMolar mass: {self.M:.2f} [g/mol]\n"
        rvalue += "\nThermophysical properties:\n"
        for key in self.__properties:
            prop_name = key.replace("_"+type(self).__name__, "")
            info = getattr(self, prop_name+"_info")(print_info=False, n_tab=1)
            rvalue += info + "\n"
        return rvalue

    def __repr__(self) -> str:
        rvalue = f"{type(self).__name__}(T={self.T:.2f}, p={self.p:.2f}, "
        for key in self.__properties:
            attr_value = getattr(self, key)
            if attr_value < 1e-2:
                rvalue += f"{key}={attr_value:.2e}, "
            else:
                rvalue += f"{key}={attr_value:.2f}, "
        rvalue = rvalue[:-2]
        rvalue += ")"
        return rvalue



"""Module with the definition of the *diffusivity*
property objects for *bismuth*."""
from typing import List
import numpy as np
from scipy.constants import atm
from scipy.constants import R
from ..tch_common_interface import OxygenDiffusivityInterface
from ..._decorators import range_warning


class BismuthOxygenDiffusivityInterface(OxygenDiffusivityInterface):
    """
    Liquid bismuth *Oxygen diffusivity* property abstract class.
    """
    @property
    def name(self) -> str:
        """
        str : Name of the property
        """
        return "o_dif"

    @property
    def units(self) -> str:
        """
        str : Oxygen diffusivity unit
        """
        return "[m^2.s^-1]"

    @property
    def long_name(self) -> str:
        """
        str : Oxygen diffusivity long name
        """
        return "oxygen diffusivity"

    @property
    def description(self) -> str:
        """
        str : Oxygen diffusivity description
        """
        return f"{self.long_name} in liquid bismuth"


class OxygenDiffusivityFitzner1980(BismuthOxygenDiffusivityInterface):
    """
    Liquid bismuth *Oxygen diffusivity* property class
    implementing the correlation by *fitzner1980*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen diffusivity* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            diffusivity in :math:`[m^2 / s]`
        """
        return np.exp(-49229 / R / T) * 1.07e-6

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "fitzner1980"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen diffusivity
        correlation function
        """
        return [951, 1100]


class OxygenDiffusivityHeshmatpour1981(BismuthOxygenDiffusivityInterface):
    """
    Liquid bismuth *Oxygen diffusivity* property class
    implementing the correlation by *heshmatpour1981*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen diffusivity* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            diffusivity in :math:`[m^2 / s]`
        """
        return np.exp(-26610 / R / T) * 1.98e-8

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "heshmatpour1981"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen diffusivity
        correlation function
        """
        return [1023, 1273]



# This test is used to check initialization from properties
import unittest
import sys
import os
import inspect
import numpy as np
sys.path.insert(0, os.path.abspath('..'))
from lbh15.properties.interface import PropertyInterface
from lbh15 import Bismuth
from scipy.constants import convert_temperature


def load_prop(module_name):
    propertyObjectList = []

    def is_valid(obj):
        return inspect.isclass(obj) and obj is not PropertyInterface \
            and not inspect.isabstract(obj) \
            and issubclass(obj, PropertyInterface)

    for _, obj in inspect.getmembers(sys.modules[module_name], is_valid):
        propertyObjectList.append(obj())
    return propertyObjectList


tol = 8
Ts = [275] + list(range(300, 1600, 50))
Ts = list(map(lambda x: convert_temperature(x, 'C', 'K'), Ts))
bismuthPs = []
for T in Ts:
    bismuthPs.append(Bismuth(T=T))


class BismuthTester(unittest.TestCase):

    def test_init_fromX(self):
        for bismuthP in bismuthPs:
            properties = load_prop('lbh15.properties.bismuth_properties')
            for prop in properties:
                name = prop.name
                if name in Bismuth.roots_to_use().keys():
                    continue
                val = getattr(bismuthP, name)
                init_dict = {name: val}
                fromX = Bismuth(**init_dict)
                self.assertAlmostEqual(bismuthP.T, fromX.T, tol, name+" FAILED")


class BismuthSolubilityTester(unittest.TestCase):

    def test_init_fromX(self):
        for bismuthP in bismuthPs:
            properties = load_prop('lbh15.properties.bismuth_thermochemical_properties.solubility_in_bismuth')
            for prop in properties:
                name = prop.name
                val = getattr(bismuthP, name)
                init_dict = {name: val}
                fromX = Bismuth(**init_dict)
                self.assertAlmostEqual(bismuthP.T, fromX.T, tol, name+" FAILED")


class BismuthDiffusivityTester(unittest.TestCase):

    def test_init_fromX(self):
        for bismuthP in bismuthPs:
            properties = load_prop('lbh15.properties.bismuth_thermochemical_properties.diffusivity_in_bismuth')
            for prop in properties:
                name = prop.name
                val = getattr(bismuthP, name)
                init_dict = {name: val}
                fromX = Bismuth(**init_dict)
                self.assertAlmostEqual(bismuthP.T, fromX.T, tol, name+" FAILED")


class BismuthThermochemicalTester(unittest.TestCase):

    def test_init_fromX(self):
        for bismuthP in bismuthPs:
            properties = load_prop('lbh15.properties.bismuth_thermochemical_properties.bismuth_thermochemical')
            for prop in properties:
                name = prop.name
                val = getattr(bismuthP, name)
                init_dict = {name: val}
                fromX = Bismuth(**init_dict)
                self.assertAlmostEqual(bismuthP.T, fromX.T, tol, name+" FAILED")


if __name__ == "__main__":
    unittest.main()



"""Module with the definition of the *solubility*
property objects for *lead-bismuth eutectic* (*lbe*)."""
from typing import List
from typing import Union
import numpy as np
from scipy.constants import atm
from ..tch_common_interface import IronSolubilityInterface
from ..tch_common_interface import NickelSolubilityInterface
from ..tch_common_interface import ChromiumSolubilityInterface
from ..tch_common_interface import OxygenSolubilityInterface
from ..._decorators import range_warning


class LBEIronSolubilityInterface(IronSolubilityInterface):
    """
    Liquid lbe *Iron solubility* property abstract class.
    """
    @property
    def description(self) -> str:
        """
        str : Iron solubility description
        """
        return f"{self.long_name} in liquid lbe"


class IronSolubilityGosse2014(LBEIronSolubilityInterface):
    """
    Liquid lbe *Iron solubility* property class
    implementing the correlation by *gosse2014*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Iron solubility* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            solubility in :math:`[wt.\\%]`
        """
        return np.power(10, 2.00-4399/T)

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "gosse2014"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Iron solubility
        correlation function
        """
        return [399.0, 1173.0]


class IronSolubilityWeeks1969(LBEIronSolubilityInterface):
    """
    Liquid lbe *Iron solubility* property class
    implementing the correlation by *weeks1969*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Iron solubility* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            solubility in :math:`[wt.\\%]`
        """
        return np.power(10, 1.85-4164/T)

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "weeks1969"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Iron solubility
        correlation function
        """
        return [823.0, 1053.0]


class LBENickelSolubilityInterface(NickelSolubilityInterface):
    """
    Liquid lbe *Nickel solubility* property abstract class.
    """
    @property
    def name(self) -> str:
        """
        str : Name of the property
        """
        return "ni_sol"

    @property
    def units(self) -> str:
        """
        str : Nickel solubility unit
        """
        return "[wt.%]"

    @property
    def long_name(self) -> str:
        """
        str : Nickel solubility long name
        """
        return "nickel solubility"

    @property
    def description(self) -> str:
        """
        str : Nickel solubility description
        """
        return f"{self.long_name} in liquid lbe"


class NickelSolubilityMartinelli2010(LBENickelSolubilityInterface):
    """
    Liquid lbe *Nickel solubility* property class
    implementing the correlation by *martinelli2010*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> Union[float, np.ndarray]:
        """
        Returns the value of the *Nickel solubility* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            solubility in :math:`[wt.\\%]`
        """
        return np.where(T <= 712, np.power(10, 5.2-3500/T),
                        np.power(10, 1.7-1009/T))[()]

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "martinelli2010"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Nickel solubility
        correlation function
        """
        return [603.0, 1173.0]


class NickelSolubilityGosse2014(LBENickelSolubilityInterface):
    """
    Liquid lbe *Nickel solubility* property class
    implementing the correlation by *gosse2014*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> Union[float, np.ndarray]:
        """
        Returns the value of the *Nickel solubility* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            solubility in :math:`[wt.\\%]`
        """
        return np.where(T <= 742, np.power(10, 4.32-2933/T),
                        np.power(10, 1.74-1006/T))[()]

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "gosse2014"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Nickel solubility
        correlation function
        """
        return [528.0, 1173.0]


class LBEChromiumSolubilityInterface(ChromiumSolubilityInterface):
    """
    Liquid lbe *Chromium solubility* property abstract class.
    """
    @property
    def name(self) -> str:
        """
        str : Name of the property
        """
        return "cr_sol"

    @property
    def units(self) -> str:
        """
        str : Chromium solubility unit
        """
        return "[wt.%]"

    @property
    def long_name(self) -> str:
        """
        str : Chromium solubility long name
        """
        return "cr solubility"

    @property
    def description(self) -> str:
        """
        str : Chromium solubility description
        """
        return f"{self.long_name} in liquid lbe"


class ChromiumSolubilityGosse2014(LBEChromiumSolubilityInterface):
    """
    Liquid lbe *Chromium solubility* property class
    implementing the correlation by *gosse2014*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Chromium solubility* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            solubility in :math:`[wt.\\%]`
        """
        return np.power(10, 1.12-3056/T)

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "gosse2014"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Chromium solubility
        correlation function
        """
        return [399.0, 1173.0]


class ChromiumSolubilityCourouau2004(LBEChromiumSolubilityInterface):
    """
    Liquid lbe *Chromium solubility* property class
    implementing the correlation by *courouau2004*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Chromium solubility* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            solubility in :math:`[wt.\\%]`
        """
        return np.power(10, 1.07-3022/T)

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "courouau2004"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Chromium solubility
        correlation function
        """
        return [643.0, 813.0]


class ChromiumSolubilityMartynov1998(LBEChromiumSolubilityInterface):
    """
    Liquid lbe *Chromium solubility* property class
    implementing the correlation by *martynov1998*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Chromium solubility* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            solubility in :math:`[wt.\\%]`
        """
        return np.power(10, -0.02-2280/T)

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "martynov1998"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Chromium solubility
        correlation function
        """
        return [673.0, 773.0]


class OxygenSolubility(OxygenSolubilityInterface):
    """
    Liquid lbe *Oxygen solubility* property class.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen solubility* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            solubility in :math:`[wt.\\%]`
        """
        return np.power(10, 2.25-4125/T)

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen solubility
        correlation function
        """
        return [673.0, 1013.0]

    @property
    def description(self) -> str:
        """
        str : Oxygen solubility description
        """
        return f"{self.long_name} in liquid lbe"



"""Module with the definition of the *solubility*
property objects for *lead*."""
from typing import List
import numpy as np
from scipy.constants import atm
from lbh15.properties.interface import PropertyInterface
from ..tch_common_interface import IronSolubilityInterface
from ..tch_common_interface import NickelSolubilityInterface
from ..tch_common_interface import ChromiumSolubilityInterface
from ..tch_common_interface import OxygenSolubilityInterface
from ..._decorators import range_warning


class IronSolubility(IronSolubilityInterface):
    """
    Liquid lead *Iron solubility* property class.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Iron solubility* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            solubility in :math:`[wt.\\%]`
        """
        return np.power(10, 2.11-5225/T)

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "gosse2014"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Iron solubility
        correlation function
        """
        return [600.0, 1173.0]

    @property
    def description(self) -> str:
        """
        str : Iron solubility description
        """
        return f"{self.long_name} in liquid lead"


class NickelSolubility(NickelSolubilityInterface):
    """
    Liquid lead *Nickel solubility* property class.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Nickel solubility* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            solubility in :math:`[wt.\\%]`
        """
        return np.power(10, 1.36-1395/T)

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "gosse2014"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Nickel solubility
        correlation function
        """
        return [598.0, 917.0]

    @property
    def description(self) -> str:
        """
        str : Nickel solubility description
        """
        return f"{self.long_name} in liquid lead"


class LeadChromiumSolubilityInterface(ChromiumSolubilityInterface):
    """
    Liquid lead *Chromium solubility* property abstract class.
    """
    @property
    def description(self) -> str:
        """
        str : Chromium solubility description
        """
        return f"{self.long_name} in liquid lead"


class ChromiumSolubilityAlden1958(LeadChromiumSolubilityInterface):
    """
    Liquid lead *Chromium solubility* property class
    implementing the correlation by *alden1958*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Chromium solubility* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            solubility in :math:`[wt.\\%]`
        """
        return np.power(10, 3.74-6750/T)

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "alden1958"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Chromium solubility
        correlation function
        """
        return [1181.0, 1483.0]


class ChromiumSolubilityVenkatraman1988(LeadChromiumSolubilityInterface):
    """
    Liquid lead *Chromium solubility* property class
    implementing the correlation by *venkatraman1988*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Chromium solubility* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            solubility in :math:`[wt.\\%]`
        """
        return np.power(10, 3.7-6720/T)

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "venkatraman1988"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Chromium solubility
        correlation function
        """
        return [1173.0, 1473.0]


class ChromiumSolubilityGosse2014(LeadChromiumSolubilityInterface):
    """
    Liquid lead *Chromium solubility* property class
    implementing the correlation by *gosse2014*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Chromium solubility* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            solubility in :math:`[wt.\\%]`
        """
        return np.power(10, 3.62-6648/T)

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "gosse2014"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Chromium solubility
        correlation function
        """
        return [601.0, 1773.0]


class SiliconSolubility(PropertyInterface):
    """
    Liquid lead *Silicon solubility* property class.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Silicon solubility* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            solubility in :math:`[wt.\\%]`
        """
        return np.power(10, 3.886-7180/T)

    @property
    def name(self) -> str:
        """
        str : Name of the property
        """
        return "si_sol"

    @property
    def units(self) -> str:
        """
        str : Silicon solubility unit
        """
        return "[wt.%]"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Silicon solubility
        correlation function
        """
        return [1323.0, 1523.0]

    @property
    def long_name(self) -> str:
        """
        str : Silicon solubility long name
        """
        return "silicon solubility"

    @property
    def description(self) -> str:
        """
        str : Silicon solubility description
        """
        return f"{self.long_name} in liquid lead"


class OxygenSolubility(OxygenSolubilityInterface):
    """
    Liquid lead *Oxygen solubility* property class.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen solubility* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            solubility in :math:`[wt.\\%]`
        """
        return np.power(10, 3.23-5043/T)

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen solubility
        correlation function
        """
        return [673.0, 1373.0]

    @property
    def description(self) -> str:
        """
        str : Oxygen solubility description
        """
        return f"{self.long_name} in liquid lead"



"""Module with the definition of the *thermo-chemical*
property objects for *lead*."""
from typing import List
from typing import Union
import numpy as np
from scipy.constants import atm
from scipy.constants import R
from ..tch_common_interface import OxygenPartialPressureInterface
from ..tch_common_interface import MolarEnthalpyInterface
from ..tch_common_interface import MolarEntropyInterface
from ..tch_common_interface import GibbsFreeEnergyInterface
from ..lead_properties import h
from ..._commons import LEAD_MELTING_TEMPERATURE as T_m0
from ..._commons import LEAD_MOLAR_MASS as M
from ..._commons import OXYGEN_MOLAR_MASS as M_O
from ..._decorators import range_warning


class LeadOxygenPartialPressureInterface(OxygenPartialPressureInterface):
    """
    *Oxygen partial pressure in liquid lead divided by the
    Oxygen concentration in liquid lead squared* property abstract class.
    """
    def initialization_helper(self,
                              property_value: float) -> Union[None, float]:
        """
        Returns the temperature guess value according to the value
        of the Oxygen partial pressure in liquid lead divided by the
        Oxygen concentration in liquid lead squared passed as argument.
        It is used by the root finder algorithm.

        Parameters
        ----------
        property_value : float
            oxygen partial pressure in liquid lead divided by the
            oxygen concentration in liquid lead squared
            in :math:`[Pa /wt.\\%^2]`

        Returns
        -------
        float
            Temperature guess value in :math:`[K]`
        """
        return 1600

    @property
    def description(self) -> str:
        """
        str : Oxygen partial pressure in liquid lead divided by the
        Oxygen concentration in liquid lead squared description
        """
        return f"{self.long_name} in liquid lead"


class OxygenPartialPressureOtsuka1979(LeadOxygenPartialPressureInterface):
    """
    *Oxygen partial pressure in liquid lead divided by the
    Oxygen concentration in liquid lead squared* property class
    implementing the correlation by *otsuka1979*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen partial pressure in liquid lead
        divided by the Oxygen concentration in liquid lead squared* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            Oxygen partial pressure divided by Oxygen concentration
            squared in :math:`[Pa / wt.\\%^2]`
        """
        return np.power(10, 2 / 2.3 / R * (-118600 / T + 14.1))\
            * M * M / M_O / M_O * 101325

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "otsuka1979"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen partial
        pressure divided by the Oxygen concentration squared
        correlation function
        """
        return [1073, 1673]


class OxygenPartialPressureOtsuka1981(LeadOxygenPartialPressureInterface):
    """
    *Oxygen partial pressure in liquid lead divided by the
    Oxygen concentration in liquid lead squared* property class
    implementing the correlation by *otsuka1981*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen partial pressure in liquid lead
        divided by the Oxygen concentration in liquid lead squared* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            Oxygen partial pressure divided by Oxygen concentration
            squared in :math:`[Pa / wt.\\%^2]`
        """
        return np.power(10, 2 / 2.3 / R * (-117170 / T + 12.9))\
            * M * M / M_O / M_O * 101325

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "otsuka1981"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen partial
        pressure divided by the Oxygen concentration squared
        correlation function
        """
        return [1023, 1273]


class OxygenPartialPressureGanesan2006(LeadOxygenPartialPressureInterface):
    """
    *Oxygen partial pressure in liquid lead divided by the
    Oxygen concentration in liquid lead squared* property class
    implementing the correlation by *ganesan2006*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen partial pressure in liquid lead
        divided by the Oxygen concentration in liquid lead squared* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            Oxygen partial pressure divided by Oxygen concentration
            squared in :math:`[Pa / wt.\\%^2]`
        """
        return np.power(10, 2 / 2.3 / R * (-121349 / T + 16.906))\
            * M * M / M_O / M_O * 101325

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "ganesan2006"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen partial
        pressure divided by the Oxygen concentration squared
        correlation function
        """
        return [815, 1090]


class OxygenPartialPressureAlcock1964(LeadOxygenPartialPressureInterface):
    """
    *Oxygen partial pressure in liquid lead divided by the
    Oxygen concentration in liquid lead squared* property class
    implementing the correlation by *alcock1964*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen partial pressure in liquid lead
        divided by the Oxygen concentration in liquid lead squared* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            Oxygen partial pressure divided by Oxygen concentration
            squared in :math:`[Pa / wt.\\%^2]`
        """
        return np.power(10, 2 / 2.3 / R * (-119411 / T + 12.222))\
            * M * M / M_O / M_O * 101325

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "alcock1964"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen partial
        pressure divided by the Oxygen concentration squared
        correlation function
        """
        return [783, 973]


class OxygenPartialPressureSzwarc1972(LeadOxygenPartialPressureInterface):
    """
    *Oxygen partial pressure in liquid lead divided by the
    Oxygen concentration in liquid lead squared* property class
    implementing the correlation by *szwarc1972*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen partial pressure in liquid lead
        divided by the Oxygen concentration in liquid lead squared* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            Oxygen partial pressure divided by Oxygen concentration
            squared in :math:`[Pa / wt.\\%^2]`
        """
        return np.power(10, 2 / 2.3 / R * (-105855 / T + 18.661))\
            * M * M / M_O / M_O * 101325

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "szwarc1972"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen partial
        pressure divided by the Oxygen concentration squared
        correlation function
        """
        return [1012, 1353]


class OxygenPartialPressureCharle1976(LeadOxygenPartialPressureInterface):
    """
    *Oxygen partial pressure in liquid lead divided by the
    Oxygen concentration in liquid lead squared* property class
    implementing the correlation by *charle1976*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen partial pressure in liquid lead
        divided by the Oxygen concentration in liquid lead squared* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            Oxygen partial pressure divided by Oxygen concentration
            squared in :math:`[Pa / wt.\\%^2]`
        """
        return np.power(10, 2 / 2.3 / R * (-119840 / T + 15.794))\
            * M * M / M_O / M_O * 101325

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "charle1976"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen partial
        pressure divided by the Oxygen concentration squared
        correlation function
        """
        return [1173, 1373]


class OxygenPartialPressureIsecke1977(LeadOxygenPartialPressureInterface):
    """
    *Oxygen partial pressure in liquid lead divided by the
    Oxygen concentration in liquid lead squared* property class
    implementing the correlation by *isecke1977*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen partial pressure in liquid lead
        divided by the Oxygen concentration in liquid lead squared* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            Oxygen partial pressure divided by Oxygen concentration
            squared in :math:`[Pa / wt.\\%^2]`
        """
        return np.power(10, 2 / 2.3 / R * (-120376 / T + 16.255))\
            * M * M / M_O / M_O * 101325

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "isecke1977"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen partial
        pressure divided by the Oxygen concentration squared
        correlation function
        """
        return [1173, 1373]


class OxygenPartialPressureTaskinen1979(LeadOxygenPartialPressureInterface):
    """
    *Oxygen partial pressure in liquid lead divided by the
    Oxygen concentration in liquid lead squared* property class
    implementing the correlation by *taskinen1979*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen partial pressure in liquid lead
        divided by the Oxygen concentration in liquid lead squared* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            Oxygen partial pressure divided by Oxygen concentration
            squared in :math:`[Pa / wt.\\%^2]`
        """
        return np.power(10, 2 / 2.3 / R * (-116717 / T + 12.699))\
            * M * M / M_O / M_O * 101325

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "taskinen1979"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen partial
        pressure divided by the Oxygen concentration squared
        correlation function
        """
        return [1073, 1203]


class OxygenPartialPressureFisher1966(LeadOxygenPartialPressureInterface):
    """
    *Oxygen partial pressure in liquid lead divided by the
    Oxygen concentration in liquid lead squared* property class
    implementing the correlation by *fisher1966*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen partial pressure in liquid lead
        divided by the Oxygen concentration in liquid lead squared* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            Oxygen partial pressure divided by Oxygen concentration
            squared in :math:`[Pa / wt.\\%^2]`
        """
        return np.power(10, 2 / 2.3 / R * (-106395 / T + 10.254))\
            * M * M / M_O / M_O * 101325

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "fisher1966"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen partial
        pressure divided by the Oxygen concentration squared
        correlation function
        """
        return [903, 1253]


class MolarEnthalpy(MolarEnthalpyInterface):
    """
    Liquid lead *molar enthalpy variation* property class.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *molar enthalpy variation* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            molar enthalpy in :math:`[J/mol]`
        """
        return h().correlation(T, p) * M / 1000

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the molar enthalpy
        variation correlation function
        """
        return [T_m0, 2000]

    @property
    def description(self) -> str:
        """
        str : Molar enthalpy variation description
        """
        return f"{self.long_name} in liquid lead"


class MolarEntropy(MolarEntropyInterface):
    """
    Liquid lead *molar entropy variation* property class.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *molar entropy variation* by
        applying the property correlation. Specific heat capacity correlation
        by *sobolev2011* is adopted.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            molar entropy in :math:`[J/(mol \\cdot K)]`
        """
        return M / 1000 * (176.2 * np.log(T / T_m0)
                           - 4.923e-2 * (T - T_m0)
                           + 7.72e-6 * (T * T - T_m0 * T_m0)
                           + 7.62e5 * (1 / T / T - 1 / T_m0 / T_m0))

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the molar entropy
        variation correlation function
        """
        return [T_m0, 2000]

    @property
    def description(self) -> str:
        """
        str : Molar entropy variation description
        """
        return f"{self.long_name} in liquid lead"


class GibbsFreeEnergy(GibbsFreeEnergyInterface):
    """
    Liquid lead *Gibbs free energy variation* property class.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Gibbs free energy variation* by
        applying the property correlation. Specific heat capacity correlation
        by *sobolev2011* is adopted.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            Gibbs free energy in :math:`[J/mol]`
        """
        return MolarEnthalpy().correlation(T, p)\
            - T * MolarEntropy().correlation(T, p)

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Gibbs free energy
        variation correlation function
        """
        return [T_m0, 2000]

    @property
    def description(self) -> str:
        """
        str : Gibbs free energy description
        """
        return f"{self.long_name} in liquid lead"



"""Module with the definition of decorators supporting the,
implementation of package functions"""
import collections.abc
import inspect
import warnings
from functools import wraps
import numpy as np

warnings.simplefilter("always")


def range_warning(function):
    """
    Decorator used to check validity range
    of correlation
    """
    @wraps(function)
    def wrapper(*args):
        range_lim = args[0].range
        p_name = args[0].long_name
        temp = args[1]
        if hasattr(temp, "__len__"):
            temp = temp[0]
        if temp < range_lim[0] or temp > range_lim[1]:
            if (len(args) == 4) and args[3]:
                warnings.warn(f"The {p_name} is requested at "
                              f"temperature value of {temp:.2f} K "
                              "that is not in validity range "
                              f"[{range_lim[0]:.2f}, {range_lim[1]:.2f}] K",
                              stacklevel=3)
        return function(*args)
    return wrapper
